@@ -1,3 +1,17 @@
+# Copyright (C) 2022-2026  Hsin Yuan Yeh <iapyeh@gmail.com>
+#
+# This file is part of Sshscript.
+#
+# Sshscript is free software; you can redistribute it and/or modify it under the
+# terms of the MIT License.
+#
+# Sshscript is distributed in the hope that it will be useful, but WITHOUT ANY
+# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
+# A PARTICULAR PURPOSE.  See the MIT License for more details.
+#
+# You should have received a copy of the MIT License along with Sshscript;
+# if not, write to the Free Software Foundation, Inc.,
+# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA.
 import __main__
 import time,subprocess
 import threading, os, sys, re
@@ -9,11 +23,33 @@ try:
 except ImportError:
     from .sshscripterror import SSHScriptError
 loop = asyncio.get_event_loop()
+
+class WithChannelWrapper(object):
+    def __init__(self,channel):
+        self.c = channel
+    def sendline(self,s='\n',secondsToWaitResponse=1):
+        return self.c.sendline(s,secondsToWaitResponse)
+    def expect(self,rawpat,timeout=60,stdout=True,stderr=True):
+        return self.c.expect(rawpat,timeout,stdout,stderr)
+    def expectStderr(self,rawpat,timeout=60):
+        return self.c.expect(rawpat,timeout,False,True)
+    def expectStdout(self,rawpat,timeout=60):
+        return self.c.expect(rawpat,timeout,True,False)
+    def wait(self,seconds=None):
+        return self.c.wait(seconds)
+
+    @property
+    def stdout(self):
+        return self.c.stdout
+    @property
+    def stderr(self):
+        return self.c.stderr
+
+
 class GenericChannel(object):
     def __init__(self):
         self.allStdoutBuf = []
         self.allStderrBuf = [] 
-        self.lock = threading.Lock()
         self._stdout = ''
         self._stderr = ''        
         self.stdoutBuf = []
@@ -23,16 +59,15 @@ class GenericChannel(object):
         self._lastIOTime = time.time()
 
         # dump-related
-        if os.environ.get('VERBOSE') or os.environ.get('DEBUG'):
+        self.stdoutDumpBuf = []
+        self.stderrDumpBuf = []
+        if os.environ.get('VERBOSE'):
             self.dump2sys = True
-            self.stdoutPrefix = os.environ.get('VERBOSE_STDOUT_PREFIX','▏').encode('utf8')
-            self.stderrPrefix = os.environ.get('VERBOSE_STDERR_PREFIX','🐞').encode('utf8')
-            self.stdoutDumpBuf = []
-            self.stderrDumpBuf = []
+            #🔵🔴🟠🟡🟢🟣🟤⭕⬜⬛🔲🟦🟥🟧🟨🟩🟪🟫🛑🔶🔷🔸🔹🔺🔻 
+            self.stdoutPrefix = os.environ.get('VERBOSE_STDOUT_PREFIX','🟩').encode('utf8')
+            self.stderrPrefix = os.environ.get('VERBOSE_STDERR_PREFIX','🟨').encode('utf8')
         else:
-            #self.dump2sys = os.environ.get('VERBOSE',sys.stdout.isatty())
             self.dump2sys = False
-
     @property
     def stdout(self):
         return self._stdout
@@ -48,7 +83,6 @@ class GenericChannel(object):
         self.lock.acquire()
         del self.stdoutBuf[:]
         del self.stderrBuf[:]
-        self.lock.release()    
 
         if self.dump2sys:
             if len(self.stdoutDumpBuf):
@@ -58,6 +92,10 @@ class GenericChannel(object):
                 sys.stderr.buffer.write(self.stderrPrefix + b''.join(self.stderrDumpBuf) + b'\n')
                 sys.stderr.buffer.flush()
 
+        del self.stdoutDumpBuf[:]
+        del self.stderrDumpBuf[:]
+        self.lock.release()    
+
         if self._stderr and self.owner.sshscript._paranoid:
             self.close()
             raise SSHScriptError(self.stderr)
@@ -66,9 +104,6 @@ class GenericChannel(object):
         if waitingInterval is None:
             waitingInterval = float(os.environ.get('CMD_INTERVAL',0.5))
         remain = waitingInterval - (time.time() - self._lastIOTime)
-        #如果有螢幕輸出的化，幫助它順序維持正確的順序
-        #sys.stdout.flush()
-        #sys.stderr.flush()
         while remain > 0:
             await asyncio.sleep(remain)    
             remain = waitingInterval - (time.time() - self._lastIOTime)
@@ -76,55 +111,63 @@ class GenericChannel(object):
         sys.stdout.flush()
         sys.stderr.flush()
 
-    def expect(self,pat,timeout=60,stdout=True,stderr=True):
+    def expect(self,rawpat,timeout=60,stdout=True,stderr=True):
+        # rawpat:bytes,str,re.Pattern or list of them
+
         # this is a blocking function
-        # 只會搜尋目前還buffred的資料，如果要搜尋新收到的資料，
-        # 須在執行命令(seld.sendline())前先呼叫 self.recv()
-        data = [lambda:self._stdout, lambda:self._stderr]
+        def checkStdout():
+            return self._stdout + (b''.join(self.stdoutBuf)).decode('utf8')
+        def checkStderr():
+            return self._stderr + (b''.join(self.stderrBuf)).decode('utf8')            
+
+        #data = [checkStdout, checkStderr]
         targets = []
         if stdout:
-            targets.append(0)
+            targets.append(checkStdout)
         if stderr:
-            targets.append(1)
-        if isinstance(pat,bytes):
-            pat = re.compile(pat.decode('utf8'),re.I)
-        elif isinstance(pat,str):
-            pat = re.compile(pat,re.I)
-        elif isinstance(pat,re.Pattern):
-            pass
-        else:
-            raise ValueError('expect() only accept str or re.Pattern')
+            targets.append(checkStderr)
+        
+        pats = []
+        if not (isinstance(rawpat,list) or isinstance(rawpat,tuple)):
+            rawpat = [rawpat]
+        for pat in rawpat:
+            if isinstance(pat,bytes):
+                pat = re.compile(pat.decode('utf8'),re.I)
+            elif isinstance(pat,str):
+                pat = re.compile(pat,re.I)
+            elif isinstance(pat,re.Pattern):
+                pass
+            else:
+                raise ValueError('expect() only accept bytes,str,re.Pattern or list of them')
+            pats.append(pat)
+
         async def _wait():
             startTime = time.time()
             while True:
                 if time.time() - startTime > timeout:
                     raise TimeoutError(f'Not found: {pat} ' + '\n')
-                for i in targets:
-                    if pat.search(data[i]()):
-                        await self.waitio(1)
-                        return
+                for dataSource in targets:
+                    di = dataSource()
+                    for pat in pats:
+                        if pat.search(di):
+                            await self.waitio(1)
+                            return
                 await asyncio.sleep(0.1)
         return loop.run_until_complete(_wait())    
-    
-    def expectStderr(self,s,timeout=60):
-        return self.expect(s,timeout,stdout=False,stderr=True)
-
-    def expectStdout(self,s,timeout=60):
-        return self.expect(s,timeout,stdout=True,stderr=False)
-    
+        
     def wait(self,seconds=None):
         async def _wait(seconds):
             await self.waitio(seconds)
         return loop.run_until_complete(_wait(seconds))                
 
     def __enter__(self):
-        return self
+        return WithChannelWrapper(self)
     
     def __exit__(self,*args):
         self.wait(1)
         self.close()
     
-    def sendline(self,s,secondsToWaitResponse=1):
+    def sendline(self,s='\n',secondsToWaitResponse=1):
         __main__.SSHScript.logger.debug(f'sendline: {s}')
         if not s[-1] == '\n': s += '\n'
         # 確保跟前面的一個指令有點「距離」，不要在還在接收資料時送出下一個指令
@@ -134,26 +177,28 @@ class GenericChannel(object):
             self.recv(secondsToWaitResponse)
         return n
 
-    def addStdoutData(self,x):       
-        self._lastIOTime = time.time()
-        if not x: return
+    def addStdoutData(self,x):  
         # 不要讓self.lock.acquire()影響self._lastIOTime
+        self._lastIOTime = time.time()
+        if not x:return
         self.lock.acquire()
         self.stdoutBuf.append(x)
         self.allStdoutBuf.append(x)
-        self.lock.release()
+        lines = None
         if self.dump2sys:
             if b'\n' in x:
                 lines = ((b''.join(self.stdoutDumpBuf))+x).split(b'\n')
-                for line in lines[:-1]:
-                    sys.stdout.buffer.write(self.stdoutPrefix + line + b'\n')
-                self.stdoutDumpBuf = [lines[-1]]
-                sys.stdout.buffer.flush()
+                lastItem = lines.pop()
+                del self.stdoutDumpBuf[:]
+                if lastItem:
+                    self.stdoutDumpBuf.append(lastItem)
             else:
-                #print('add',[x])
                 self.stdoutDumpBuf.append(x)
-
-        self._lastIOTime = time.time()
+        self.lock.release()
+        if lines:
+            for line in lines:
+                sys.stdout.buffer.write(self.stdoutPrefix + line + b'\n')
+            sys.stdout.buffer.flush()
     
     def addStderrData(self,x):
         # 不要讓self.lock.acquire()影響self._lastIOTime
@@ -162,17 +207,21 @@ class GenericChannel(object):
         self.lock.acquire()
         self.stderrBuf.append(x)
         self.allStderrBuf.append(x)
-        self.lock.release()
+        lines = None
         if self.dump2sys:
             if b'\n' in x:
                 lines = ((b''.join(self.stderrDumpBuf))+x).split(b'\n')
-                for line in lines[:-1]:
-                    sys.stderr.buffer.write(self.stderrPrefix + line + b'\n')
-                self.stderrDumpBuf = [lines[-1]]
-                sys.stderr.buffer.flush()
+                lastItem = lines.pop()
+                del self.stderrDumpBuf[:]
+                if lastItem:
+                    self.stderrDumpBuf.append(lastItem)
             else:
                 self.stderrDumpBuf.append(x)
-        self._lastIOTime = time.time()
+        self.lock.release()
+        if lines:
+            for line in lines:
+                sys.stderr.buffer.write(self.stderrPrefix + line + b'\n')
+            sys.stderr.buffer.flush()
 
 class POpenPipeChannel(GenericChannel):
     def __init__(self,owner,cp,timeout):
@@ -202,9 +251,9 @@ class POpenPipeChannel(GenericChannel):
             self.masterFd[0]: self.addStdoutData, #stdout
             self.masterFd[1]: self.addStderrData  #stderr
         }
-        while buffer:           
+        while buffer :           
             try:
-                for fd in select(buffer, [], [])[0]:
+                for fd in select(buffer, [], [],1)[0]:
                     try:
                         # diff of read and read1:
                         # See: https://stackoverflow.com/questions/57726771/what-the-difference-between-read-and-read1-in-python
@@ -230,10 +279,6 @@ class POpenPipeChannel(GenericChannel):
                     break
                 else:
                     raise
-        #if len(stdoutBuffer):
-        #    self.addStdoutData(b''.join(stdoutBuffer))
-        #if len(stderrBuffer):
-        #    self.addStdoutData(b''.join(stderrBuffer))
 
     def close(self):
         self.wait(1) # at least 1 seconds has no io
@@ -241,6 +286,7 @@ class POpenPipeChannel(GenericChannel):
             self.masterFd[0].close()
             self.masterFd[1].close()
             self.slaveFd[0].close()
+            self.cp.stdin.close()
 
             while self.cp.poll() is None:
                 try:
@@ -267,11 +313,13 @@ class POpenChannel(GenericChannel):
             threading.Thread(target=self._reading).start()
 
     def send(self,s):
-        self._lastIOTime = time.time()
+        #self._lastIOTime = time.time()
         b = s.encode('utf-8')
         # write to popen's stdin
-        os.write(self.masterFd[0],b)
-        os.fsync(self.masterFd[0])
+        #os.write(self.masterFd[0],b)
+        #os.fsync(self.masterFd[0])
+        self.cp.stdin.write(b)
+        self.cp.stdin.flush()
         self._lastIOTime = time.time()
     
     def _reading(self):
@@ -280,9 +328,9 @@ class POpenChannel(GenericChannel):
             self.masterFd[0]: self.addStdoutData, #stdout
             self.masterFd[1]: self.addStderrData  #stderr
         }
-        while buffer:           
+        while buffer:
             try:
-                for fd in select(buffer, [], [])[0]:
+                for fd in select(buffer, [], [],1)[0]:
                     try:
                         data = os.read(fd, 1024) # read available
                         
@@ -308,10 +356,6 @@ class POpenChannel(GenericChannel):
                     break
                 else:
                     raise
-        #if len(stdoutBuffer):
-        #    self.addStdoutData(b''.join(stdoutBuffer))
-        #if len(stderrBuffer):
-        #    self.addStdoutData(b''.join(stderrBuffer))
 
     def close(self):
         self.wait(1) # at least 1 seconds has no io
@@ -320,7 +364,7 @@ class POpenChannel(GenericChannel):
             os.close(self.masterFd[1])
             os.close(self.slaveFd[0])
             os.close(self.slaveFd[1])
-
+            self.cp.stdin.close()
             while self.cp.poll() is None:
                 try:
                     self.cp.wait(self.timeout)
